// Copyright (c) 2014-2022 NetEase, Inc.
// All right reserved.

import 'dart:async';

typedef VoidResult = NEResult<void>;

class NEResult<T> {
  final int code;

  final String? msg;

  final T? data;

  const NEResult({required this.code, this.msg, this.data});

  const NEResult.success(): this(code: 0);

  NEResult.successWith(T data): this(code: 0, data: data);

  bool isSuccess() => code == 0 || code == 200;

  T get nonNullData => data!;

  @override
  String toString() {
    if (isSuccess()) {
      return 'Success($data)';
    } else {
      return 'Failure($code, $msg)';
    }
  }
}

extension CommonResultExtension<T> on NEResult<T> {

  NEResult<T> onFailure(void Function(int, String?) action) {
    if (!isSuccess()) {
      action(code, msg);
    }
    return this;
  }
}

extension ObjectResultExtension<T extends Object> on NEResult<T> {

  NEResult<U> map<U>(U Function(T) mapper) =>
      isSuccess() ? NEResult<U>(
        code: code,
        msg: msg,
        data: mapper(nonNullData),
      ) : NEResult<U>(code: code, msg: msg);

  NEResult<T> onSuccess(void Function(T) action) {
    if (isSuccess()) {
      action(nonNullData);
    }
    return this;
  }
}

extension VoidResultExtension on NEResult<void> {

  NEResult<U> map<U>(U Function() mapper) =>
      isSuccess() ? NEResult<U>(
        code: code,
        msg: msg,
        data: mapper(),
      ) : NEResult<U>(code: code, msg: msg);

  NEResult<void> onSuccess(void Function() action) {
    if (isSuccess()) {
      action();
    }
    return this;
  }
}

extension CommonFutureResultExtension<T> on Future<NEResult<T>> {

  Future<NEResult<T>> onFailure(FutureOr Function(int, String?) action) async {
    final result = await this;
    if (!result.isSuccess()) {
      await action(result.code, result.msg);
    }
    return this;
  }
}

extension ObjectFutureResultExtension<T extends Object> on Future<NEResult<T>> {

  Future<NEResult<T>> onSuccess(FutureOr Function(T) action) async {
    final result = await this;
    if (result.isSuccess()) {
      await action(result.nonNullData);
    }
    return this;
  }

  Future<NEResult<U>> map<U>(FutureOr Function(T) mapper) async {
    final first = await this;
    if (first.isSuccess()) {
      final next = await mapper(first.nonNullData);
      try {
        if (next is NEResult) {
          return next as NEResult<U>;
        } else {
          return NEResult<U>.successWith(next as U);
        }
      } catch (e, _) {
        assert((){
          throw e;
        }());
        return NEResult<U>(code: -1, msg: e.toString());
      }
    } else {
      return NEResult<U>(code: first.code, msg: first.msg);
    }
  }
}

extension VoidFutureResultExtension on Future<NEResult<void>> {

  Future<NEResult<void>> onSuccess(FutureOr Function() action) async {
    final result = await this;
    if (result.isSuccess()) {
      await action();
    }
    return this;
  }

  Future<NEResult<U>> map<U>(FutureOr Function() mapper) async {
    final first = await this;
    if (first.isSuccess()) {
      final next = await mapper();
      try {
        if (next is NEResult) {
          return next as NEResult<U>;
        } else {
          return NEResult<U>.successWith(next as U);
        }
      } catch (e, _) {
        assert((){
          throw e;
        }());
        return NEResult<U>(code: -1, msg: e.toString());
      }
    } else {
      return NEResult<U>(code: first.code, msg: first.msg);
    }
  }
}